/*
 ==============================================================================
 
 the ResFS1 object. A 2nd order controllable resonator
 (2 poles 2 zeros digital resonator - Forced solution)
 
 implementation is Forced Solution I
 
 version 1, feb 2011, T.V.
 modif mars 2011, Ch.Gondre
 modif mars 2012
 
 version 1.1 : octobre 2013 Ch.G. ajout dsp64.
 compil windows & osx pour max 5 et 6, y compris 64 bits
 
 1rst inlet : (signal) excitation
 2nd inlet  : (signal/float) resonnance frequency
 2rd inlet  : (signal/float) decay rate
 4th inlet  : (signal/float) gain
 
 pole radius is R = exp(-alpha/fs)
 
 normalisation : (1 - R*R) / R
 
 ============================================================================
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can use,
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS, and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided
 * only with a limited warranty and the software's author, the holder of
 * the economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards
 * their requirements in conditions enabling the security of their
 * systems and/or data to be ensured and, more generally, to use and
 * operate it in the same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 
 ============================================================================
 */

// Sur windows, il faut definir ca si compilation avec le max sdk 6.0
// peut etre encore avec le 6.1.3 ?
#ifndef DENORM_WANT_FIX
#define DENORM_WANT_FIX		(1)
#endif

#include "ext.h"        // standard Max include, always required
#include "ext_obex.h"   // required for new style objects
#include "z_dsp.h"      // required for MSP objects
#include <math.h>

#define S2M_RESFS1_VERSION "1.1"

//==============================================================================

void *ResFS1_class;

typedef struct _ResFS1
{
    t_pxobject l_obj;

    double fs;           // sample rate
    double TwoPiOverSR;  // 2.* pi / fs
    double f0;           // resonnance freq
    double alpha;        // decay rate
    double R;
    double gain;         // gain given by the user
    double Norm;         // gain pour la normalisation
    
    double Xnm1;
    double Xn;
    double Ynm1;
    double Yn;
    double Enm1;
    
    double omegaR; // used only if there's no signal connected to the freq inlet
    double x1;
    double y1;

    short freqConnected;    // is a signal connected to the freq inlet
    short dampingConnected; // is a signal connected to the damping inlet
    short gainConnected;    // is a signal connected to the gain inlet
    
} t_ResFS1;

//==============================================================================

void ResFS1_dsp (t_ResFS1 *x, t_signal **sp, short *count);
void ResFS1_dsp64 (t_ResFS1 *x, t_object *dsp64, short *count, double samplerate, long maxvectorsize, long flags);

// on a 4 fonctions pour le process, selon que la freq et le damping
// sont des signaux ou non. Le test pour le gain (signal ou pas)
// est fait au sein de ces fonctions.
t_int *ResFS1_freqDampingPerform (t_int *w);
t_int *ResFS1_freqPerform (t_int *w);
t_int *ResFS1_dampingPerform (t_int *w);
t_int *ResFS1_basicPerform (t_int *w);
// 64 versions :
void ResFS1_freqDampingPerform64 (t_ResFS1 *x, t_object *dsp64, double **ins, long numins, double **outs, long numouts, long sampleframes, long flags, void *userparam);
void ResFS1_freqPerform64 (t_ResFS1 *x, t_object *dsp64, double **ins, long numins, double **outs, long numouts, long sampleframes, long flags, void *userparam);
void ResFS1_dampingPerform64 (t_ResFS1 *x, t_object *dsp64, double **ins, long numins, double **outs, long numouts, long sampleframes, long flags, void *userparam);
void ResFS1_basicPerform64 (t_ResFS1 *x, t_object *dsp64, double **ins, long numins, double **outs, long numouts, long sampleframes, long flags, void *userparam);

void *ResFS1_new (double f0, double decayrate, double gain);
void ResFS1_assist (t_ResFS1 *x, void *b, long m, long a, char *s);
void ResFS1_int (t_ResFS1 *x, long n);
void ResFS1_float (t_ResFS1 *x, double value);
void ResFS1_clear (t_ResFS1 *x);

//==============================================================================

int C74_EXPORT main (void)
{
    t_class *c = class_new ("s2m.resFS1~", (method) ResFS1_new, (method) dsp_free,
            sizeof (t_ResFS1), 0L, A_DEFFLOAT, A_DEFFLOAT, A_DEFFLOAT, 0);
    
    class_addmethod (c, (method) ResFS1_dsp, "dsp", A_CANT, 0);
    class_addmethod (c, (method) ResFS1_dsp64, "dsp64", A_CANT, 0);
    class_addmethod (c, (method) ResFS1_assist, "assist", A_CANT, 0);
    class_addmethod (c, (method) ResFS1_clear, "clear", 0);
    class_addmethod (c, (method) ResFS1_int, "int", A_LONG, 0);
    class_addmethod (c, (method) ResFS1_float, "float", A_FLOAT, 0);
    class_dspinit (c);
    class_register (CLASS_BOX, c);
    ResFS1_class = c;

    post ("[s2m.resFS1~] version %s, Th. Voinier & Ch. Gondre - cnrs lma s2m - %s", S2M_RESFS1_VERSION, __DATE__);

    return(0);
}

//==============================================================================

void *ResFS1_new (double f0, double decayrate, double gain)
{
    t_ResFS1 *x = NULL;
    
    if ((x = (t_ResFS1 *) object_alloc (ResFS1_class)))
    {
        // 4 inlets for : excitation, f0, damping, gain
        dsp_setup ((t_pxobject *) x, 4);
        
        // initializations using the arguments :
        x->f0 = fabs (f0);
        
        // we do not allow negative decay rate
        x->alpha = decayrate >= 0. ? decayrate : 0.;
        
        x->fs = sys_getsr();
        x->R = exp (-x->alpha / x->fs);
        x->gain = gain;
        
        x->Norm = (1. - x->R * x->R) / x->R;
        
        x->TwoPiOverSR = TWOPI / x->fs;
        
        x->omegaR = x->f0 * x->TwoPiOverSR;
        x->x1 = x->R * cos (x->omegaR);
        x->y1 = x->R * sin (x->omegaR);
        
        ResFS1_clear (x); // set sample memories to 0.

        outlet_new ((t_object *) x, "signal"); // one signal outlet
        x->l_obj.z_misc = Z_NO_INPLACE;
    }
    return (x);
}

//==============================================================================

/** Setting up assistance strings. */
void ResFS1_assist (t_ResFS1 *x, void *b, long m, long a, char *s)
{
    if (m == ASSIST_OUTLET)
    {
        sprintf (s,"(signal) Output");
    }
    else
    {
        switch (a)
        {
            case 0:
                sprintf (s,"(signal) Excitation Input");
                break;
            case 1:
                sprintf (s,"(signal/float) Resonance Frequency");
                break;
            case 2:
                sprintf (s,"(signal/float) Damping");
                break;
            case 3:
                sprintf (s,"(signal/float) Gain");
                break;
        }
    }
}

//==============================================================================
//                      MESSAGES FROM MAX                                     //
//==============================================================================
/** brief clear sample memory to recover from blowup. */
void ResFS1_clear (t_ResFS1 *x)
{
    x->Ynm1 = x->Yn = 0.;
    x->Xnm1 = x->Xn = 0.;
}

void ResFS1_int (t_ResFS1 *x, long n)
{
    ResFS1_float (x, (double) n);
}

void ResFS1_float (t_ResFS1 *x, double val)
{
    long in = proxy_getinlet ((t_object *) x);
    
    if (in == 3 )// gain inlet
    {
        x->gain = val;
        x->Norm = (1 - x->R * x->R) / x->R;
        return;
    }
    
    if (in == 1) // freq inlet
    {         
        x->f0 = fabs (val);
    }
    else if (in == 2) // decayrate inlet
    {
        // we do not allow negative decay rate
        x->alpha = val >= 0. ? val : 0.;
        x->R = exp (-x->alpha / x->fs);
        x->Norm = (1. - x->R * x->R) / x->R;
    }
    
    x->omegaR = x->f0 * x->TwoPiOverSR;
    x->x1 = x->R * cos (x->omegaR);
    x->y1 = x->R * sin (x->omegaR);
}

//==============================================================================

void ResFS1_dsp (t_ResFS1 *x, t_signal **sp, short *count)
{
    if (x->fs != sp[0]->s_sr)
    {
        x->fs = sp[0]->s_sr;
        x->R = exp (-x->alpha/x->fs);
        x->Norm = (1. - x->R * x->R) / x->R;
        x->TwoPiOverSR = TWOPI / x->fs;
        x->omegaR = x->f0 * x->TwoPiOverSR;
        x->x1 = x->R * cos (x->omegaR);
        x->y1 = x->R * sin (x->omegaR);
    }
    
    ResFS1_clear(x);
    
    x->freqConnected = count[1];    // signal connected to the frequency inlet?
    x->dampingConnected = count[2]; // signal connected to the damping inlet?
    x->gainConnected = count[3];    // signal connected to the gain inlet?
    
    // on a 4 fonctions pour le process, selon que la freq et le damping
    // sont des signaux ou non. Le test pour le gain (signal ou pas)
    // sera fait au sein de ces fonctions.
    
    if ( x->freqConnected && x->dampingConnected ) // frequence ET damping en signal
    {
        dsp_add (ResFS1_freqDampingPerform, 7, sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec, sp[3]->s_vec, sp[4]->s_vec, x, sp[0]->s_n);
    }
    else if ( x->freqConnected ) // frequence en signal
    {
        dsp_add (ResFS1_freqPerform, 6, sp[0]->s_vec, sp[1]->s_vec, sp[3]->s_vec, sp[4]->s_vec, x, sp[0]->s_n);
    }
    else if (x->dampingConnected) // damping en signal
    {
        dsp_add (ResFS1_dampingPerform, 6, sp[0]->s_vec, sp[2]->s_vec, sp[3]->s_vec, sp[4]->s_vec, x, sp[0]->s_n);
    }
    else // Eventuellement le gain en signal, mais rien d'autre
    {
        dsp_add (ResFS1_basicPerform, 5, sp[0]->s_vec, sp[3]->s_vec, sp[4]->s_vec, x, sp[0]->s_n);
    }
}

void ResFS1_dsp64 (t_ResFS1 *x, t_object *dsp64, short *count, double samplerate, long maxvectorsize, long flags)
{
    if (x->fs != samplerate)
    {
        x->fs = samplerate;
        x->R = exp (-x->alpha / samplerate);
        x->Norm = (1. - x->R * x->R) / x->R;
        x->TwoPiOverSR = TWOPI / x->fs;
        x->omegaR = x->f0 * x->TwoPiOverSR;
        x->x1 = x->R * cos (x->omegaR);
        x->y1 = x->R * sin (x->omegaR);
    }
    
    ResFS1_clear(x);
    
    x->freqConnected = count[1];       // signal connected to the frequency inlet?
    x->dampingConnected = count[2];    // signal connected to the damping inlet?
    x->gainConnected = count[3];       // signal connected to the gain inlet?
    
    // on a 4 fonctions pour le process, selon que la freq et le damping
    // sont des signaux ou non. Le test pour le gain (signal ou pas)
    // sera fait au sein de ces fonctions.
    
    if ( x->freqConnected && x->dampingConnected ) // frequence ET damping en signal
    {
        object_method (dsp64, gensym ("dsp_add64"), x, (t_perfroutine64) ResFS1_freqDampingPerform64, 0, NULL);
    }
    else if ( x->freqConnected ) // frequence en signal
    {
        object_method (dsp64, gensym ("dsp_add64"), x, (t_perfroutine64) ResFS1_freqPerform64, 0, NULL);
    }
    else if (x->dampingConnected) // damping en signal
    {
        object_method (dsp64, gensym ("dsp_add64"), x, (t_perfroutine64) ResFS1_dampingPerform64, 0, NULL);
    }
    else // Eventuellement le gain en signal, mais rien d'autre
    {
        object_method (dsp64, gensym ("dsp_add64"), x, (t_perfroutine64) ResFS1_basicPerform64, 0, NULL);
    }
}

/*==============================================================================

                      PERFORM ROUTINES                                      

      on a 4 fonctions pour le process, selon que la freq et le damping
      sont des signaux ou non. Le test pour le gain (signal ou pas)
         est fait au sein de ces fonctions.
==============================================================================*/

/** this perform routine is used when signals are connected to BOTH the
 frequency inlet and the damping inlet */
t_int *ResFS1_freqDampingPerform (t_int *w)
{
    // assign from parameters
    t_float *in1  = (t_float *) w[1];  // excitation signal
    t_float *inFreq = (t_float *) w[2];   // freq signal
    t_float *inDecay = (t_float *) w[3];   // decayrate signal
    t_float *inGain = (t_float *) w[4];   // gain signal
    t_float *out1 = (t_float *) w[5];  // out signal

    t_ResFS1  *x = (t_ResFS1 *) w[6];
    int n = (int) w[7];                // vectorsize

    double R             = x->R;
    double TwoPiOverSR   = x->TwoPiOverSR;

    double Xnm1          = x->Xnm1;
    double Xn            = x->Xn;
    double Ynm1          = x->Ynm1;
    double Yn            = x->Yn;
    double Enm1          = x->Enm1;
    double f0            = x->f0;
    double gain          = x->gain;
    double alpha         = x->alpha;
    
    double previousF0 = f0;
    double previousAlpha = alpha;

    double En;
    
    double x1            = x->x1;
    double y1            = x->y1;
    double omegaR        = x->omegaR;
    double Norm          = x->Norm;
    double scale;
	double decay;
    
    float fs            = x->fs;
    bool testBool       = false;
    
    if (x->l_obj.z_disabled)
        goto out;
    
    // si le l'entree gain est un scalaire
    if (!x->gainConnected)
    {
        scale = Norm * gain ;
    
        while (n--)
        {
            // recompute some coefficients if needed
            f0 = *inFreq++;
            decay = *inDecay++;
            alpha = decay >= 0. ? decay : 0.;
            
            if (f0 != previousF0)
            {
                previousF0 = f0;
                omegaR  = f0 * TwoPiOverSR;
                testBool = true;
            }
            
            if (alpha != previousAlpha)
            {
                previousAlpha = alpha;
                R = exp (-alpha / fs);
                Norm = (1. - R * R) / R;
                scale = Norm * gain;
                testBool = true;
            }
            
            if (testBool)
            {
                x1 = R * cos (omegaR);
                y1 = R * sin (omegaR);
                testBool = false;
            }
            
            En = scale * *in1++;
            
            // difference equation
            Xn  = x1 * Xnm1 - y1 * Ynm1 + Enm1;
            Yn  = y1 * Xnm1 + x1 * Ynm1;
            
#ifdef DENORM_WANT_FIX
            FIX_DENORM_NAN_DOUBLE(Yn);
            FIX_DENORM_NAN_DOUBLE(Xn);
#endif
            *out1++ = Yn;
        
            // prepare next iteration
            Ynm1 = Yn;        
            Xnm1 = Xn;
            Enm1 = En;
        }
    }
    else // si l'entree gain est un signal
    {        
        while (n--)
        {
            // recompute some coefficients if needed
            f0 = *inFreq++;
            decay = *inDecay++;
            alpha = decay >= 0. ? decay : 0.;
            
            if (f0 != previousF0)
            {
                previousF0 = f0;
                omegaR  = f0 * TwoPiOverSR;
                testBool = true;
            }
            
            if (alpha != previousAlpha)
            {
                previousAlpha = alpha;
                R = exp (-alpha / fs);
                Norm = (1. - R * R) / R;
                testBool = true;
            }
            
            if (testBool)
            {
                x1 = R * cos (omegaR);
                y1 = R * sin (omegaR);
                testBool = false;
            }

            gain = *inGain++;
            En = gain * Norm * *in1++;            

            // difference equation
            Xn  = x1 * Xnm1 - y1 * Ynm1 + Enm1;
            Yn  = y1 * Xnm1 + x1 * Ynm1;
            
#ifdef DENORM_WANT_FIX
            FIX_DENORM_NAN_DOUBLE(Yn);
            FIX_DENORM_NAN_DOUBLE(Xn);
#endif
            *out1++ = Yn;
        
            // prepare next iteration
            Ynm1 = Yn;        
            Xnm1 = Xn;
            Enm1 = En;
        }
        x->gain = gain;
    }
    
    x->x1 = x1;
    x->y1 = y1;
    x->f0 = f0;
    x->omegaR = omegaR; 
    x->Norm = Norm;
    x->R = R;
    x->alpha = alpha;
        
    x->Ynm1 = Ynm1;
    x->Yn   = Yn;
    x->Xnm1 = Xnm1;
    x->Xn   = Xn;
    x->Enm1 = Enm1;
    
out:
    return (w + 8);
}

/* 64 version. this perform routine is used when signals are connected to BOTH
* the frequency inlet and the damping inlet */
void ResFS1_freqDampingPerform64 (t_ResFS1 *x, t_object *dsp64, double **ins, long numins, double **outs, long numouts, long sampleframes, long flags, void *userparam)
{
    // assign from parameters
    t_double *in1 = ins[0];     // excitation signal
    t_double *inFreq = ins[1];  // freq signal
    t_double *inDecay = ins[2]; // decayrate signal
    t_double *inGain = ins[3];  // gain signal
    t_double *out1 = outs[0];   // out signal
    
    double R             = x->R;
    double TwoPiOverSR   = x->TwoPiOverSR;
    
    double Xnm1          = x->Xnm1;
    double Xn            = x->Xn;
    double Ynm1          = x->Ynm1;
    double Yn            = x->Yn;
    double Enm1          = x->Enm1;
    double f0            = x->f0;
    double gain          = x->gain;
    double alpha         = x->alpha;
    
    double previousF0 = f0;
    double previousAlpha = alpha;
    
    double En;
    
    double x1            = x->x1;
    double y1            = x->y1;
    double omegaR        = x->omegaR;
    double Norm          = x->Norm;
    double scale;
	double decay;
    
    float fs            = x->fs;
    bool testBool       = false;
    
    if (x->l_obj.z_disabled)
        return;
    
    // si le l'entree gain est un scalaire
    if (!x->gainConnected)
    {
        scale = Norm * gain ;
        
        while (sampleframes--)
        {
            // recompute some coefficients if needed
            f0 = *inFreq++;
            decay = *inDecay++;
            alpha = decay >= 0. ? decay : 0.;
            
            if (f0 != previousF0)
            {
                previousF0 = f0;
                omegaR  = f0 * TwoPiOverSR;
                testBool = true;
            }
            
            if (alpha != previousAlpha)
            {
                previousAlpha = alpha;
                R = exp (-alpha / fs);
                Norm = (1. - R * R) / R;
                scale = Norm * gain;
                testBool = true;
            }
            if (testBool)
            {
                x1 = R * cos (omegaR);
                y1 = R * sin (omegaR);
                testBool = false;
            }
            
            En = scale * *in1++;
            
            // difference equation
            Xn  = x1 * Xnm1 - y1 * Ynm1 + Enm1;
            Yn  = y1 * Xnm1 + x1 * Ynm1;
            
#ifdef DENORM_WANT_FIX
            FIX_DENORM_NAN_DOUBLE(Yn);
            FIX_DENORM_NAN_DOUBLE(Xn);
#endif
            *out1++ = Yn;
            
            // prepare next iteration
            Ynm1 = Yn;
            Xnm1 = Xn;
            Enm1 = En;
        }
    }
    else // si l'entree gain est un signal
    {
        while (sampleframes--)
        {
            // recompute some coefficients
            f0 = *inFreq++;
            decay = *inDecay++;
            alpha = decay >= 0. ? decay : 0.;
            
            if (f0 != previousF0)
            {
                previousF0 = f0;
                omegaR  = f0 * TwoPiOverSR;
                testBool = true;
            }
            
            if (alpha != previousAlpha)
            {
                previousAlpha = alpha;
                R = exp (-alpha / fs);
                Norm = (1. - R * R) / R;
                testBool = true;
            }
            
            if (testBool)
            {
                x1 = R * cos (omegaR);
                y1 = R * sin (omegaR);
                testBool = false;
            }
            
            gain = *inGain++;
            En = gain * Norm * *in1++;
            
            // difference equation
            Xn  = x1 * Xnm1 - y1 * Ynm1 + Enm1;
            Yn  = y1 * Xnm1 + x1 * Ynm1;
            
#ifdef DENORM_WANT_FIX
            FIX_DENORM_NAN_DOUBLE(Yn);
            FIX_DENORM_NAN_DOUBLE(Xn);
#endif
            *out1++ = Yn;
            
            // prepare next iteration
            Ynm1 = Yn;
            Xnm1 = Xn;
            Enm1 = En;
        }
        x->gain = gain;
    }
    
    
    x->x1 = x1;
    x->y1 = y1;
    x->f0 = f0;
    x->omegaR = omegaR;
    x->Norm = Norm;
    x->R = R;
    x->alpha = alpha;
    
    x->Ynm1 = Ynm1;
    x->Yn   = Yn;
    x->Xnm1 = Xnm1;
    x->Xn   = Xn;
    x->Enm1 = Enm1;
}

//----------------------------------------------------------------------------//


/** this perform routine is used when a signal is connected to the freq inlet. */
t_int *ResFS1_freqPerform (t_int *w)
{
    // assign from parameters
    t_float *in1 = (t_float *) w[1];    // excitation signal
    t_float *inFreq = (t_float *) w[2]; // freq signal
    t_float *inGain = (t_float *) w[3]; // gain signal
    t_float *out1 = (t_float *) w[4];   // out signal

    t_ResFS1  *x = (t_ResFS1 *) w[5];
    int n = (int) w[6]; // vectorsize

    double R             = x->R;
    double TwoPiOverSR   = x->TwoPiOverSR;

    double Xnm1          = x->Xnm1;
    double Xn            = x->Xn;
    double Ynm1          = x->Ynm1;
    double Yn            = x->Yn;
    double Enm1          = x->Enm1;
    double f0            = x->f0;
    double previousF0    = f0;
    double gain          = x->gain;
    
    double En;
    
    double x1            = x->x1;
    double y1            = x->y1;
    double omegaR        = x->omegaR;
    double Norm          = x->Norm;
    double scale;
    
    if (x->l_obj.z_disabled)
        goto out;
    
    // si le l'entree gain est un scalaire
    if (!x->gainConnected)
    {
        scale = Norm * gain ;
    
        while (n--)
        {
            En = scale * *in1++;
            // recompute some coefficients
            f0 = *inFreq++;
            
            // do we need to recompute coefficients?
            if (f0 != previousF0)
            {
                previousF0 = f0;
                omegaR  = f0 * TwoPiOverSR;
                x1 = R * cos (omegaR);
                y1 = R * sin (omegaR);
            }
            
            // difference equation
            Xn  = x1 * Xnm1 - y1 * Ynm1 + Enm1;
            Yn  = y1 * Xnm1 + x1 * Ynm1;
            
#ifdef DENORM_WANT_FIX
            FIX_DENORM_NAN_DOUBLE(Yn);
            FIX_DENORM_NAN_DOUBLE(Xn);
#endif
            *out1++ = Yn;
        
            // prepare next iteration
            Ynm1 = Yn;        
            Xnm1 = Xn;
            Enm1 = En;
            
        }
    }
    else    // si l'entree gain est un signal
    {        
        while (n--)
        {
            En = gain * Norm * *in1++;
            // recompute some coefficients
            f0 = *inFreq++;
            
            gain = *inGain++;
            
            // do we need to recompute coefficients?
            if (f0 != previousF0)
            {
                previousF0 = f0;
                omegaR  = f0 * TwoPiOverSR;
                x1 = R * cos (omegaR);
                y1 = R * sin (omegaR);
            }

            // difference equation
            Xn  = x1 * Xnm1 - y1 * Ynm1 + Enm1;
            Yn  = y1 * Xnm1 + x1 * Ynm1;
            
#ifdef DENORM_WANT_FIX
            FIX_DENORM_NAN_DOUBLE(Yn);
            FIX_DENORM_NAN_DOUBLE(Xn);
#endif
            *out1++ = Yn;
        
            // prepare next iteration
            Ynm1 = Yn;        
            Xnm1 = Xn;
            Enm1 = En;
            
        }
        x->gain = gain;
    }

    x->x1 = x1;
    x->y1 = y1;
    x->f0 = f0;
    x->omegaR = omegaR; 
        
    x->Ynm1 = Ynm1;
    x->Yn   = Yn;
    x->Xnm1 = Xnm1;
    x->Xn   = Xn;
    x->Enm1 = Enm1;
    
out:
    return (w + 7);
}

/** 64 version. this perform routine is used when a signal is connected to the freq inlet */
void ResFS1_freqPerform64 (t_ResFS1 *x, t_object *dsp64, double **ins, long numins, double **outs, long numouts, long sampleframes, long flags, void *userparam)
{
    // assign from parameters
    t_double *in1 = ins[0];     // excitation signal
    t_double *inFreq = ins[1];  // freq signal
    t_double *inGain = ins[2];  // gain signal
    t_double *out1 = outs[0];   // out signal
    
    double R             = x->R;
    double TwoPiOverSR   = x->TwoPiOverSR;
    
    double Xnm1          = x->Xnm1;
    double Xn            = x->Xn;
    double Ynm1          = x->Ynm1;
    double Yn            = x->Yn;
    double Enm1          = x->Enm1;
    double f0            = x->f0;
    double previousF0    = f0;
    double gain          = x->gain;
    
    double En;
    
    double x1            = x->x1;
    double y1            = x->y1;
    double omegaR        = x->omegaR;
    double Norm          = x->Norm;
    double scale;
    
    if (x->l_obj.z_disabled)
        return;
    
    // si le l'entree gain est un scalaire
    if (!x->gainConnected)
    {
        scale = Norm * gain ;
        
        while (sampleframes--)
        {
            En = scale * *in1++;
            f0 = *inFreq++;
            
            // recompute some coefficients if needed
            if (f0 != previousF0)
            {
                previousF0 = f0;
                omegaR  = f0 * TwoPiOverSR;
                x1 = R * cos (omegaR);
                y1 = R * sin (omegaR);
            }
            
            // difference equation
            Xn  = x1 * Xnm1 - y1 * Ynm1 + Enm1;
            Yn  = y1 * Xnm1 + x1 * Ynm1;
            
#ifdef DENORM_WANT_FIX
            FIX_DENORM_NAN_DOUBLE(Yn);
            FIX_DENORM_NAN_DOUBLE(Xn);
#endif
            *out1++ = Yn;
            
            // prepare next iteration
            Ynm1 = Yn;
            Xnm1 = Xn;
            Enm1 = En;
            
        }
    }
    else // si l'entree gain est un signal
    {
        while (sampleframes--)
        {
            En = gain * Norm * *in1++;
            f0 = *inFreq++;
            gain = *inGain++;
            
            // recompute some coefficients if needed
            if (f0 != previousF0)
            {
                previousF0 = f0;
                omegaR  = f0 * TwoPiOverSR;
                x1 = R * cos (omegaR);
                y1 = R * sin (omegaR);
            }
            
            // difference equation
            Xn  = x1 * Xnm1 - y1 * Ynm1 + Enm1;
            Yn  = y1 * Xnm1 + x1 * Ynm1;
            
#ifdef DENORM_WANT_FIX
            FIX_DENORM_NAN_DOUBLE(Yn);
            FIX_DENORM_NAN_DOUBLE(Xn);
#endif
            *out1++ = Yn;
            
            // prepare next iteration
            Ynm1 = Yn;
            Xnm1 = Xn;
            Enm1 = En;
        }
        x->gain = gain;
    }
    
    x->x1 = x1;
    x->y1 = y1;
    x->f0 = f0;
    x->omegaR = omegaR;
    
    x->Ynm1 = Ynm1;
    x->Yn   = Yn;
    x->Xnm1 = Xnm1;
    x->Xn   = Xn;
    x->Enm1 = Enm1;
}
//----------------------------------------------------------------------------//


/** this perform routine is used when a signal is connected to the damping inlet */
t_int *ResFS1_dampingPerform (t_int *w)
{
    // assign from parameters
    t_float *in1 = (t_float *) w[1];        // excitation signal
    t_float *inDecay = (t_float *) w[2];    // decayrate signal
    t_float *inGain = (t_float *) w[3];     // gain signal
    t_float *out1 = (t_float *) w[4];       // out signal

    t_ResFS1  *x = (t_ResFS1 *) w[5];
    int n = (int) w[6]; // vectorsize

    double R     = x->R;
    double Xnm1  = x->Xnm1;
    double Xn    = x->Xn;
    double Ynm1  = x->Ynm1;
    double Yn    = x->Yn;
    double Enm1  = x->Enm1;
    double gain  = x->gain;
    double alpha = x->alpha;
    
    double previousAlpha = alpha;

    double En;
    
    double x1       = x->x1;
    double y1       = x->y1;
    double omegaR   = x->omegaR;
    double Norm     = x->Norm;
    double scale;
    
    double fs       = x->fs;
    
    if (x->l_obj.z_disabled)
        goto out;
    
    // si le l'entree gain est un scalaire
    if (!x->gainConnected)
    {
        scale = Norm * gain ;
    
        while (n--)
        {
            float decay = *inDecay++;
            alpha = decay >= 0. ? decay : 0.;
            
            // recompute some coefficients if needed
            if (alpha != previousAlpha)
            {
                previousAlpha = alpha;
                R = exp (-alpha / fs);
                
                Norm = (1. - R * R) / R;
               
                scale = Norm * gain ;
                x1 = R * cos (omegaR);
                y1 = R * sin (omegaR);
            }
            
            En = scale * *in1++;

            // difference equation
            Xn  = x1 * Xnm1 - y1 * Ynm1 + Enm1;
            Yn  = y1 * Xnm1 + x1 * Ynm1;
            
#ifdef DENORM_WANT_FIX
            FIX_DENORM_NAN_DOUBLE(Yn);
            FIX_DENORM_NAN_DOUBLE(Xn);
#endif
            *out1++ = Yn;
            
            Ynm1 = Yn;        
            Xnm1 = Xn;
            Enm1 = En;
        }
    }
    else    // si l'entree gain est un signal
    {        
        while (n--)
        {
            float decay = *inDecay++;
            alpha = decay >= 0. ? decay : 0.;
        
            // recompute some coefficients if needed
            if (alpha != previousAlpha)
            {
                previousAlpha = alpha;
                R = exp (-alpha / fs);
                
                Norm = (1. - R * R) / R;
                
                scale = Norm * gain;
                x1 = R * cos (omegaR);
                y1 = R * sin (omegaR);
            }

            gain = *inGain++;
            En = gain * Norm * *in1++;            

            // difference equation
            Xn  = x1 * Xnm1 - y1 * Ynm1 + Enm1;
            Yn  = y1 * Xnm1 + x1 * Ynm1;
            
#ifdef DENORM_WANT_FIX
            FIX_DENORM_NAN_DOUBLE(Yn);
            FIX_DENORM_NAN_DOUBLE(Xn);
#endif
            *out1++ = Yn;
        
            // prepare next iteration
            Ynm1 = Yn;        
            Xnm1 = Xn;
            Enm1 = En;
        }
        x->gain = gain;
    }


    x->x1 = x1;
    x->y1 = y1;
    
    x->Norm = Norm;
    x->R = R;
    x->alpha = alpha;
    
    x->Ynm1 = Ynm1;
    x->Yn   = Yn;
    x->Xnm1 = Xnm1;
    x->Xn   = Xn;
    x->Enm1 = Enm1;
    
out:
    return (w + 7);
}


/** 64 version. this perform routine is used when a signal is connected to the damping inlet */
void ResFS1_dampingPerform64 (t_ResFS1 *x, t_object *dsp64, double **ins, long numins, double **outs, long numouts, long sampleframes, long flags, void *userparam)
{
    // assign from parameters
    t_double *in1 = ins[0];         // excitation signal
    t_double *inDecay = ins[1];     // decayrate signal
    t_double *inGain = ins[2];      // gain signal
    t_double *out1 = outs[0];       // out signal
    
    double R     = x->R;
    double Xnm1  = x->Xnm1;
    double Xn    = x->Xn;
    double Ynm1  = x->Ynm1;
    double Yn    = x->Yn;
    double Enm1  = x->Enm1;
    double gain  = x->gain;
    double alpha = x->alpha;
    
    double previousAlpha = alpha;
    
    double En;
    
    double x1       = x->x1;
    double y1       = x->y1;
    double omegaR   = x->omegaR;
    double Norm     = x->Norm;
    double scale;
    
    double fs       = x->fs;
    
    if (x->l_obj.z_disabled)
        return;
    
    // si le l'entree gain est un scalaire
    if (!x->gainConnected)
    {
        scale = Norm * gain ;
        
        while (sampleframes--)
        {
            double decay = *inDecay++;
            alpha = decay >= 0. ? decay : 0.;
            
            // do we need to recompute coefficients?
            if (alpha != previousAlpha)
            {
                previousAlpha = alpha;
                R = exp (-alpha / fs);
                
                Norm = (1. - R * R) / R;
                
                scale = Norm * gain ;
                x1 = R * cos (omegaR);
                y1 = R * sin (omegaR);
            }
            
            En = scale * *in1++;
            
            // difference equation
            Xn  = x1 * Xnm1 - y1 * Ynm1 + Enm1;
            Yn  = y1 * Xnm1 + x1 * Ynm1;
            
#ifdef DENORM_WANT_FIX
            FIX_DENORM_NAN_DOUBLE(Yn);
            FIX_DENORM_NAN_DOUBLE(Xn);
#endif
            *out1++ = Yn;
            
            // prepare next iteration
            Ynm1 = Yn;
            Xnm1 = Xn;
            Enm1 = En;
        }
    }
    else    // si l'entree gain est un signal
    {
        while (sampleframes--)
        {
            double decay = *inDecay++;
            alpha = decay >= 0. ? decay : 0.;
            
            // do we need to recompute coefficients?
            if (alpha != previousAlpha)
            {
                previousAlpha = alpha;
                R = exp (-alpha / fs);
                
                Norm = (1. - R * R) / R;
                
                scale = Norm * gain;
                x1 = R * cos (omegaR);
                y1 = R * sin (omegaR);
            }
            
            gain = *inGain++;
            En = gain * Norm * *in1++;
            
            // difference equation
            Xn  = x1 * Xnm1 - y1 * Ynm1 + Enm1;
            Yn  = y1 * Xnm1 + x1 * Ynm1;
            
#ifdef DENORM_WANT_FIX
            FIX_DENORM_NAN_DOUBLE(Yn);
            FIX_DENORM_NAN_DOUBLE(Xn);
#endif
            *out1++ = Yn;
            
            // prepare next iteration
            Ynm1 = Yn;
            Xnm1 = Xn;
            Enm1 = En;
        }
        x->gain = gain;
    }
    
    x->x1 = x1;
    x->y1 = y1;
    
    x->Norm = Norm;
    x->R = R;
    x->alpha = alpha;
    
    x->Ynm1 = Ynm1;
    x->Yn   = Yn;
    x->Xnm1 = Xnm1;
    x->Xn   = Xn;
    x->Enm1 = Enm1;
}

//----------------------------------------------------------------------------//

/* this perform routine is called when NO signal is connected to the damping
 * inlet neither to the frequency inlet. */
t_int *ResFS1_basicPerform (t_int *w)
{
    // assign from parameters
    t_float *in1  = (t_float *) w[1];   // excitation signal
    t_float *inGain = (t_float *) w[3]; // gain signal
    t_float *out1 = (t_float *) w[3];   // out signal

    t_ResFS1  *x = (t_ResFS1 *) w[4] ;
    int n = (int) w[5];                 // vectorsize

    double Xnm1          = x->Xnm1;
    double Xn            = x->Xn;
    double Ynm1          = x->Ynm1;
    double Yn            = x->Yn;
    double Enm1          = x->Enm1;
    double Norm          = x->Norm ;
    double scale;
    
    double En;
    
    double x1            = x->x1;
    double y1            = x->y1;
    
    if (x->l_obj.z_disabled)
        goto out;
    
    // si le l'entree gain est un scalaire
    if (!x->gainConnected)
    {
        scale = Norm * x->gain ;
        
        while (n--)
        {
            En = scale * *in1++;
            // difference equation
            Xn  = x1 * Xnm1 - y1 * Ynm1 + Enm1;
            Yn  = y1 * Xnm1 + x1 * Ynm1;
            
    #ifdef DENORM_WANT_FIX
            FIX_DENORM_NAN_DOUBLE(Yn);
            FIX_DENORM_NAN_DOUBLE(Xn);
    #endif
            *out1++ = Yn;
        
            // prepare next iteration
            Ynm1 = Yn;
            Xnm1 = Xn;
            Enm1 = En;
        }
    }
    else // si l'entree gain est un signal
    {
        float gain = *inGain++;
        
        while (n--)
        {
            En = gain * Norm * *in1++;
            // difference equation
            Xn  = x1 * Xnm1 - y1 * Ynm1 + Enm1;
            Yn  = y1 * Xnm1 + x1 * Ynm1;
            
    #ifdef DENORM_WANT_FIX
            FIX_DENORM_NAN_DOUBLE(Yn);
            FIX_DENORM_NAN_DOUBLE(Xn);
    #endif
            *out1++ = Yn;
        
            // prepare next iteration
            Ynm1 = Yn;
            Xnm1 = Xn;
            Enm1 = En;
        }
        x->gain = gain;
    }
    
    x->Ynm1 = Ynm1;
    x->Yn   = Yn;
    x->Xnm1 = Xnm1;
    x->Xn   = Xn;
    x->Enm1 = Enm1;
out:
    return (w + 6);
}


/* 64 version. this perform routine is called when NO signal is connected to the damping
 * inlet neither to the frequency inlet. */
void ResFS1_basicPerform64 (t_ResFS1 *x, t_object *dsp64, double **ins, long numins, double **outs, long numouts, long sampleframes, long flags, void *userparam)
{
    // assign from parameters
    t_double *in1 = ins[0];     // excitation signal
    t_double *inGain = ins[1];  // gain signal
    t_double *out1 = outs[0];   // out signal
    
    double Xnm1          = x->Xnm1;
    double Xn            = x->Xn;
    double Ynm1          = x->Ynm1;
    double Yn            = x->Yn;
    double Enm1          = x->Enm1;
    double Norm          = x->Norm ;
    double scale;
    
    double En;
    
    double x1            = x->x1;
    double y1            = x->y1;
    
    if (x->l_obj.z_disabled)
        return;
    
    // si le l'entree gain est un scalaire
    if (!x->gainConnected)
    {
        scale = Norm * x->gain ;
        
        while (sampleframes--)
        {
            En = scale * *in1++;
            // difference equation
            Xn  = x1 * Xnm1 - y1 * Ynm1 + Enm1;
            Yn  = y1 * Xnm1 + x1 * Ynm1;
            
#ifdef DENORM_WANT_FIX
            FIX_DENORM_NAN_DOUBLE(Yn);
            FIX_DENORM_NAN_DOUBLE(Xn);
#endif
            *out1++ = Yn;
            
            // prepare next iteration
            Ynm1 = Yn;
            Xnm1 = Xn;
            Enm1 = En;
        }
    }
    else // si l'entree gain est un signal
    {
        double gain = *inGain++;
        
        while (sampleframes--)
        {
            En = gain * Norm * *in1++;
            // difference equation
            Xn  = x1 * Xnm1 - y1 * Ynm1 + Enm1;
            Yn  = y1 * Xnm1 + x1 * Ynm1;
            
#ifdef DENORM_WANT_FIX
            FIX_DENORM_NAN_DOUBLE(Yn);
            FIX_DENORM_NAN_DOUBLE(Xn);
#endif
            *out1++ = Yn;
            
            // prepare next iteration
            Ynm1 = Yn;
            Xnm1 = Xn;
            Enm1 = En;
        }
        x->gain = gain;
    }
    
    x->Ynm1 = Ynm1;
    x->Yn   = Yn;
    x->Xnm1 = Xnm1;
    x->Xn   = Xn;
    x->Enm1 = Enm1;
}

//==============================================================================
